"""
Command line entry
"""
import asyncio
import threading
from os import getpid
import psutil
from .web import Webserver
from ..models import WebserverArgs
from ..framework.communicator import CommunicatorFactory


class CommandLine:
    '''Command line entry class
    '''

    def __init__(self, **kwargs):
        self.communication = 'uart'
        self.device_provider = None
        self.communicator = None
        self.supported_commands = []
        self.input_string = None
        self.current_command = None
        self._build_options(**kwargs)
        self.webserver = Webserver(**kwargs)
        self.webserver_running = False

    def listen(self):
        # find device
        '''
        Start to find device
        '''
        self.detect_device(self.device_discover_handler)

    def detect_device(self, callback):
        '''find if there is a connected device'''
        if self.communicator is None:
            self.communicator = CommunicatorFactory.create(
                self.communication, self.options)

        self.communicator.find_device(callback)

    def device_discover_handler(self, device_provider):
        '''
        Handler after device discovered
        '''
        # check if device is in bootloader
        # TODO: if in bootloader, only allow upgrade

        # TODO: if a normal device, allow other commands

        # load device provider
        self.webserver.set_communicator(self.communicator)
        self.webserver.load_device_provider(device_provider)
        # setup command
        self.device_provider = device_provider
        self.setup_command_handler()

    def setup_command_handler(self):
        '''
        Prepare command
        '''
        self.supported_commands = self.device_provider.get_command_lines()

        while True:
            token = input(">>")
            self.input_string = token.split(" ")

            if token.strip() == 'exit':
                break

            if self.webserver_running and token.strip() != 'stop':
                print("server is on-going, please stop it")
                continue

            for command in self.supported_commands:
                if command['name'] == self.input_string[0]:
                    self.current_command = command
                    eval('self.%s()' % (command['function']))
                    break
            else:
                self.help_handler()

        self.exit_handler()
        return True

    def start_webserver(self):
        '''
        Start websocket server
        '''
        # asyncio.set_event_loop(asyncio.new_event_loop())
        self.webserver.start_webserver()

    def _build_options(self, **kwargs):
        self.options = WebserverArgs(**kwargs)

    # command handler
    def help_handler(self):
        '''
        Help handler
        '''
        if len(self.supported_commands) > 0:
            print("Usage: ")
            for command in self.supported_commands:
                print(command['name'] + " : " + command['description'])
        else:
            print("No more command line.")

    def connect_handler(self):
        '''
        Connect to device, may no need it later
        '''

    def upgrade_handler(self):
        '''upgrade command is used for firmware upgrade and followed by file name
        '''
        input_args = len(self.input_string)
        if input_args == 1:
            print("Usage:")
            print("upgrade file_name")
        else:
            file_name = self.input_string[1]
            # TODO: check device is idel
            self.device_provider.upgrade_framework(file_name)
        return True

    def record_handler(self):
        '''record command is used to save the outputs into local machine
        '''
        # TODO: check device is idel
        if not self.device_provider.is_logging:
            self.device_provider.start_data_log()
        return True

    def stop_handler(self):
        '''record command is used to save the outputs into local machine
        '''
        # TODO: check device is idel
        if self.device_provider.is_logging:
            self.device_provider.stop_data_log()

        if self.webserver_running:
            self.webserver.stop()
            self.webserver_running = False
        return True

    def get_handler(self):
        '''
        Get parameter of device
        '''
        input_args = len(self.input_string)
        conf = self.device_provider.get_conf()
        input_params_properties = conf['data']['inputParams']
        select_param = None
        if (input_args == 1):
            print("Usage: get [options]")
            print("Option: ")
            i = 2
            while i < len(input_params_properties):
                print(input_params_properties[i]['argument'])
                i += 1
            return True
        else:
            i = 2
            while i < len(input_params_properties):
                select_param = input_params_properties[i]
                if (select_param['argument'] == self.input_string[1]):
                    break
                i += 1
                if (i == len(input_params_properties)):
                    print("Usage: get [options]")
                    print("Option: ")
                    i = 2
                    while i < len(input_params_properties):
                        print(input_params_properties[i]['argument'])
                        i += 1
                    return True

        param = self.device_provider.get_param(
            {'paramId': select_param['paramId']})
        print(param['data']['value'])
        return True

    def set_handler(self):
        '''
        Set parameter of device
        '''
        input_args = len(self.input_string)
        conf = self.device_provider.get_conf()
        input_params_properties = conf['data']['inputParams']
        select_param = None
        not_in_options = False
        options = []

        if input_args == 1:
            print("Usage: set <options> <values>")
            i = 2
            while i < len(input_params_properties):
                print(input_params_properties[i]['argument'])
                i += 1
            return True
        else:
            i = 2
            while i < len(input_params_properties):
                select_param = input_params_properties[i]
                if (select_param['argument'] == self.input_string[1]):
                    break
                i += 1

        if input_args == 2:
            if i == len(input_params_properties):
                print("Usage: set <options> <values>")
                i = 2
                while i < len(input_params_properties):
                    print(input_params_properties[i]['argument'])
                    i += 1
            else:
                print("Usage: set " + select_param['argument'] + " <values>")
                print("values: ")
                print(select_param['options'])
            return True

        if select_param.__contains__('options'):
            for item in select_param['options']:
                if isinstance(item, dict):
                    options.append(int(item['key']))
                else:
                    options.append(item)

        if select_param['type'] == 'int64':
            self.input_string[2] = int(self.input_string[2])

        if select_param['type'] == "char8" and self.input_string[2] not in select_param['options']:
            not_in_options = True

        if select_param['type'] == "int64" and\
                self.input_string[2] not in options:
            not_in_options = True

        if not_in_options:
            print("Usage: set " + select_param['argument'] + " <values>")
            print("values: ")
            print(select_param['options'])
            return True

        self.device_provider.set_param({
            'paramId': select_param['paramId'],
            'value': self.input_string[2]
        })

        #TODO: display a response message to user

        return True

    def save_handler(self):
        '''
        Save device configuration
        '''
        self.device_provider.save_config()
        return True

    def server_start_handler(self):
        '''
        start a websocket server
        '''
        # self.webserver.start_websocket_server()
        webserver_thread = threading.Thread(target=self.start_webserver)
        webserver_thread.start()
        self.webserver_running = True
        return True

    def exit_handler(self):
        '''
        Exit current process
        '''
        self.webserver.stop()
        self.webserver_running = False
        pid = getpid()
        process = psutil.Process(pid)
        process.kill()

    def run_handler(self):
        '''used by customers
        '''
        return True



import struct
import collections
from .open_field_parser import decode_value
# from .dmu_field_parser import decode_value

# input packet


def get_all_parameters_parser(payload, user_configuration):
    '''
    gA parser
    '''
    error = False
    data = []
    data_len = 0
    for parameter in user_configuration:
        param_id = parameter['paramId']
        param_type = parameter['type']
        name = parameter['name']

        if param_type == 'uint8' or param_type == 'int8':
            value = decode_value(
                param_type, payload[data_len:data_len + 1])
            data_len = data_len + 1
        elif param_type == 'uint16' or param_type == 'int16':
            value = decode_value(
                param_type, payload[data_len:data_len + 2])
            data_len = data_len + 2
        elif param_type == 'uint32' or param_type == 'int32' or param_type == 'float':
            value = decode_value(
                param_type, payload[data_len:data_len + 4])
            data_len = data_len + 4
        elif param_type == 'uint64' or param_type == 'int64' or param_type == 'double':
            value = decode_value(
                param_type, payload[data_len:data_len + 8])
            data_len = data_len + 8
        elif param_type == 'ip4':
            value = decode_value(
                param_type, payload[data_len:data_len + 4])
            data_len = data_len + 4
        elif param_type == 'ip6':
            value = decode_value(
                param_type, payload[data_len:data_len + 6])
            data_len = data_len + 6
        elif 'char' in param_type:
            ctype_n = param_type.replace('char', '')
            ctype_l = int(ctype_n)
            value = decode_value(
                param_type, payload[data_len:data_len + ctype_l])
            data_len = data_len + ctype_l
        else:
            print(
                "no [{0}] when unpack_input_packet".format(param_type))
            value = False
        data.append(
            {"paramId": param_id, "name": name, "value": value})

    return data, error


def get_parameters_by_block_parser(payload, user_configuration):
    '''
    gB parser
    '''
    data = []
    error = False

    start_param_id = payload[0]
    end_param_id = payload[1]
    data_len = 2

    for i in range(start_param_id, end_param_id+1, 1):
        print('param count:', end_param_id, i)
        exist_param_conf = next((param_conf for param_conf in user_configuration
                                 if param_conf['paramId'] == i), None)
        if exist_param_conf:
            param_type = exist_param_conf['type']

            if param_type == 'uint8' or param_type == 'int8':
                value = decode_value(
                    param_type, payload[data_len:data_len + 1])
                data_len = data_len + 1
            elif param_type == 'uint16' or param_type == 'int16':
                value = decode_value(
                    param_type, payload[data_len:data_len + 2])
                data_len = data_len + 2
            elif param_type == 'uint32' or param_type == 'int32' or param_type == 'float':
                value = decode_value(
                    param_type, payload[data_len:data_len + 4])
                data_len = data_len + 4
            elif param_type == 'uint64' or param_type == 'int64' or param_type == 'double':
                value = decode_value(
                    param_type, payload[data_len:data_len + 8])
                data_len = data_len + 8
            elif param_type == 'ip4':
                value = decode_value(
                    param_type, payload[data_len:data_len + 4])
                data_len = data_len + 4
            elif param_type == 'ip6':
                value = decode_value(
                    param_type, payload[data_len:data_len + 6])
                data_len = data_len + 6
            elif 'char' in param_type:
                ctype_n = param_type.replace('char', '')
                ctype_l = int(ctype_n)
                value = decode_value(
                    param_type, payload[data_len:data_len + ctype_l])
                data_len = data_len + ctype_l
            else:
                print(
                    "no [{0}] when unpack_input_packet".format(param_type))
                value = False

            data.append({
                "paramId": i,
                "name": exist_param_conf['name'],
                "value": value
            })

    return data, error


def get_parameter_parser(payload, user_configuration):
    '''
    gP Parser
    '''
    data = None
    error = False
    param_id = decode_value('uint32', payload[0:4])
    param = filter(lambda item: item['paramId'] ==
                   param_id, user_configuration)

    try:
        first_item = next(iter(param), None)
        param_value = decode_value(
            first_item['type'], payload[4:12])
        data = {"paramId": param_id,
                "name": first_item['name'], "value": param_value}
    except StopIteration:
        error = True

    return data, error


def update_parameter_parser(payload, user_configuration):
    '''
    uP parser
    '''
    error = False
    data = decode_value('uint32', payload[0:4])
    if data:
        error = True
    return data, error


def common_input_parser(payload, user_configuration):
    '''
    General input packet parser
    '''
    return payload, False


def read_eeprom_parser(payload, user_configuration=None):
    return payload[3:], False

# output packet


def common_continuous_parser(payload, configuration):
    '''
    Unpack output packet
    '''
    if configuration is None:
        return

    data = None
    is_list = 0
    length = 0
    pack_fmt = '<'
    for value in configuration['payload']:
        if value['type'] == 'float':
            pack_fmt += 'f'
            length += 4
        elif value['type'] == 'uint32':
            pack_fmt += 'I'
            length += 4
        elif value['type'] == 'int32':
            pack_fmt += 'i'
            length += 4
        elif value['type'] == 'int16':
            pack_fmt += 'h'
            length += 2
        elif value['type'] == 'uint16':
            pack_fmt += 'H'
            length += 2
        elif value['type'] == 'double':
            pack_fmt += 'd'
            length += 8
        elif value['type'] == 'int64':
            pack_fmt += 'q'
            length += 8
        elif value['type'] == 'uint64':
            pack_fmt += 'Q'
            length += 8
        elif value['type'] == 'char':
            pack_fmt += 'c'
            length += 1
        elif value['type'] == 'uchar':
            pack_fmt += 'B'
            length += 1
        elif value['type'] == 'uint8':
            pack_fmt += 'B'
            length += 1
    len_fmt = '{0}B'.format(length)

    has_list = configuration.__contains__('isList')
    if has_list:
        is_list = configuration['isList']

    if is_list == 1:
        packet_num = len(payload) // length
        data = []
        for i in range(packet_num):
            payload_c = payload[i*length:(i+1)*length]
            try:
                pack_item = struct.pack(len_fmt, *payload_c)
                item = struct.unpack(pack_fmt, pack_item)
                out = [(value['name'], item[idx])
                       for idx, value in enumerate(configuration['payload'])]
                item = collections.OrderedDict(out)
                data.append(item)
            except Exception as ex:  # pylint: disable=broad-except
                print(
                    "error happened when decode the payload, pls restart driver: {0}"
                    .format(ex))
    else:
        try:
            pack_item = struct.pack(len_fmt, *payload)
            data = struct.unpack(pack_fmt, pack_item)
            out = [(value['name'], data[idx])
                   for idx, value in enumerate(configuration['payload'])]
            data = collections.OrderedDict(out)
        except Exception as ex:  # pylint: disable=broad-except
            print(
                "error happened when decode the payload of packets, pls restart driver: {0}"
                .format(ex))

    return data


# packet handler
def match_command_handler(packet_type):
    '''
    Find the handler for specified packet
    '''
    parser_dict = {
        'gA': get_all_parameters_parser,
        'gB': get_parameters_by_block_parser,
        'gP': get_parameter_parser,
        'sC': common_input_parser,
        'uP': update_parameter_parser,
        'rD': common_input_parser,
        'RE': read_eeprom_parser,
        'WE': common_input_parser,
        'UE': common_input_parser,
        'LE': common_input_parser,
        'SR': common_input_parser,
        'JI': common_input_parser,
        'JA': common_input_parser,
        'WA': common_input_parser
    }
    return parser_dict.get(packet_type)



"""
Context
"""
from .app_logger import AppLogger


class AppContext:
    '''
    App Context
    '''
    _active_app = None
    _logger = None

    def __init__(self):
        pass

    def set_app(self, app):
        '''
        app setter
        '''
        self._active_app = app

    def get_app(self):
        '''
        app getter
        '''
        return self._active_app

    def set_logger(self, logger):
        '''
        logger setter
        '''
        self._logger = logger

    def get_logger(self):
        '''
        logger getter
        '''
        if not self._logger:
            self._logger = AppLogger(filename='default')
        return self._logger


APP_CONTEXT = AppContext()



"""
predefined params for openimu
"""
JSON_FILE_NAME = 'openimu.json'


def get_app_names():
    '''
    define openimu app type
    '''
    app_names = ['Compass',
                 'IMU',
                 'INS',
                 'Leveler',
                 'OpenIMU',
                 'VG',
                 'VG_AHRS',
                 ]
    return app_names


APP_STR = ['INS', 'VG', 'VG_AHRS', 'Compass', 'Leveler', 'IMU', 'OpenIMU']



# -*- coding: utf-8 -*
import sys
import os
import time
import datetime
import json
import threading
import requests
from azure.storage.blob import AppendBlobService
from azure.storage.blob import ContentSettings
from .utils import resource
from .configuration import get_config


class FileLoger():
    def __init__(self, device_properties):
        '''Initialize and create a CSV file
        '''
        start_time = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        self.device_properties = device_properties
        if not self.device_properties:
            print('No properties found')
            os._exit(1)

        self.root_folder = os.path.join(resource.get_executor_path(), r'data')
        if not os.path.exists(self.root_folder):
            os.mkdir(self.root_folder)
        self.output_packets = self.device_properties['userMessages']['outputPackets']
        self.log_file_rows = {}
        self.log_file_names = {}
        self.log_files_obj = {}
        self.log_files = {}
        self.user_file_name = ''  # the prefix of log file name.
        self.msgs_need_to_log = []
        self.ws = False
        # azure app.
        self.user_id = ''
        self.file_name = ''
        self.sas_token = ''
        self.db_user_access_token = ''
        # 'http://40.118.233.18:3000/'  # TODO: set a host url
        self.host_url = get_config().ANS_PLATFORM_URL

        #
        self.threads = []  # thread of receiver and paser
        self.exit_thread = False  # flag of exit threads
        self.exit_lock = threading.Lock()  # lock of exit_thread
        self.data_dict = {}  # data container
        self.data_lock = threading.Lock()  # lock of data_queue

        self.device_log_info = None

    def start_user_log(self, file_name='', ws=False):
        '''
        start log.
        return:
                0: OK
                1: exception that has started logging already.
                2: other exception.
        '''
        try:
            if len(self.log_file_rows) > 0:
                return 1  # has started logging already.

            self.ws = ws
            self.exit_thread = False
            self.user_file_name = file_name
            start_time = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
            current_path = os.path.join(self.root_folder, start_time)
            if not os.path.exists(current_path):
                os.mkdir(current_path)

            for packet in self.output_packets:
                # if 1 == packet['save2file']:
                has_save2file = packet.__contains__('save2file')
                save2file = 1
                if has_save2file:
                    save2file = packet['save2file']

                if save2file == 1:
                    self.msgs_need_to_log.append(packet['name'])

                self.log_file_rows[packet['name']] = 0
                if self.user_file_name == '':
                    self.log_file_names[packet['name']
                                        ] = packet['name'] + '.csv'
                else:
                    self.log_file_names[packet['name']] = self.user_file_name + \
                        '_' + packet['name'] + '.csv'
                self.log_files[packet['name']
                               ] = self.log_file_names[packet['name']]

                self.log_files_obj[packet['name']] = open(
                    current_path + '/' + self.log_file_names[packet['name']], 'w')

            if self.ws:
                self.get_sas_token()
                self.data_dict.clear()
                for i, (k, v) in enumerate(self.log_files.items()):  # k:pack type  v:log file name
                    self.data_dict[v] = ''
                    threading.Thread(target=self.upload_azure,
                                     args=(k, v)).start()
            return 0
        except Exception as e:
            print('Exception! File:[{0}], Line:[{1}]. Exception:{2}'.format(
                __file__, sys._getframe().f_lineno, e))
            return 2

    def stop_user_log(self):
        '''
        stop log.
        return:
                0: OK
                1: exception that driver hasn't started logging files yet.
                2: other exception.
        '''
        rev = 0
        try:
            if len(self.log_file_rows) == 0:
                return 1  # driver hasn't started logging files yet.
            for i, (k, v) in enumerate(self.log_files_obj.items()):
                v.close()
            self.log_file_rows.clear()
            self.log_file_names.clear()
            self.log_files_obj.clear()
            rev = 0
        except Exception as e:
            print(e)
            rev = 2

        if self.ws:
            time.sleep(1)
            self.exit_lock.acquire()
            self.exit_thread = True
            self.exit_lock.release()
            self.ws = False

        return rev

    def upload_azure(self, packet_type, log_file_name):
        if self.db_user_access_token == '' or self.sas_token == '':
            print(
                "Error: Can not upload log to azure since token is empty! Please check the network.")

        print(datetime.datetime.now().strftime(
            '%Y_%m_%d_%H_%M_%S:'), log_file_name, ' start.')

        config = get_config()
        accountName = config.AZURE_STORAGE_ACCOUNT
        countainerName = config.AZURE_STORAGE_DATA_CONTAINER
        url_name = datetime.datetime.now().strftime(
            '%Y_%m_%d_%H_%M_%S') + '-' + self.user_id + '-' + log_file_name
        bcreate_blob_ok = False

        error_connection = 'ConnectionError'
        error_authorization = 'AuthenticationFailed'
        ii = 0
        while True:
            # get data from data_dict.
            self.data_lock.acquire()
            text = self.data_dict[log_file_name]
            self.data_dict[log_file_name] = ''
            self.data_lock.release()

            # check if user stop logging data.
            self.exit_lock.acquire()
            if self.exit_thread:
                # check for internet and text
                if text == '' or (not self.internet_on()):
                    self.exit_lock.release()
                    break
                else:
                    pass
            self.exit_lock.release()

            # let CPU have a break.
            if text == '':
                time.sleep(1)
                continue

            # create blob on azure
            if not bcreate_blob_ok:
                try:
                    self.append_blob_service = AppendBlobService(account_name=accountName,
                                                                 sas_token=self.sas_token,
                                                                 protocol='http')
                    self.append_blob_service.create_blob(container_name=countainerName, blob_name=url_name,
                                                         content_settings=ContentSettings(content_type='text/plain'))
                    bcreate_blob_ok = True
                    threading.Thread(target=self.save_to_db_task, args=(
                        packet_type, log_file_name, url_name)).start()
                except Exception as e:
                    # print('Exception when create_blob:', type(e), e)
                    if error_connection in str(e):
                        pass
                    elif error_authorization in str(e):
                        self.get_sas_token()
                        self.append_blob_service = AppendBlobService(account_name=accountName,
                                                                     sas_token=self.sas_token,
                                                                     protocol='http')
                    print('Retry to create_blob again...')
                    continue

            # append blob on azure
            try:
                # self.append_blob_service.append_blob_from_text(countainerName, fileName, text, progress_callback=self.upload_callback)
                self.append_blob_service.append_blob_from_text(
                    countainerName, url_name, text)
            except Exception as e:
                # print('Exception when append_blob:', type(e), e)
                if error_connection in str(e):
                    pass
                elif error_authorization in str(e):
                    self.get_sas_token()
                    self.append_blob_service = AppendBlobService(account_name=accountName,
                                                                 sas_token=self.sas_token,
                                                                 protocol='http')
                    # if append blob failed, do not drop 'text', but push 'text' to data_dict and re-append next time.
                    self.data_lock.acquire()
                    self.data_dict[log_file_name] = text + \
                        self.data_dict[log_file_name]
                    self.data_lock.release()

        if bcreate_blob_ok:
            # if not self.save_to_ans_platform(packet_type, log_file_name):
            #     print('save_to_ans_platform failed.')
            print(datetime.datetime.now().strftime(
                '%Y_%m_%d_%H_%M_%S:'), log_file_name, ' done.')

    def save_to_db_task(self, packet_type, file_name, url_name):
        if not self.save_to_ans_platform(packet_type, file_name, url_name):
            print('save_to_ans_platform failed.')

    def append(self, packet_type, packet):
        if len(self.log_file_rows) == 0:  # if hasn't started logging.
            return

        if packet_type in self.msgs_need_to_log:
            self.log(packet_type, packet)

    def get_log_file_names(self):
        return self.log_file_names.copy()

    def log(self, packet_type, data):
        ''' Parse the data, read in from the unit, and generate a data file using
            the json properties file to create a header and specify the precision
            of the data in the resulting data file.
        '''
        output_packet = next(
            (x for x in self.output_packets if x['name'] == packet_type), None)

        '''Write row of CSV file based on data received.  Uses dictionary keys for column titles
        '''
        if self.log_file_rows[packet_type] == 0:
            # Loop through each item in the data dictionary and create a header from the json
            #   properties that correspond to the items in the dictionary
            labels = ''
            # for key in data:
            for i, (k, v) in enumerate(data.items()):
                '''dataStr = output_packet['payload'][i]['name'] + \
                          ' [' + \
                          output_packet['payload'][i]['unit'] + \
                          ']'''
                data_str = output_packet['payload'][i]['name']
                unit_str = output_packet['payload'][i]['unit']
                if unit_str == '':
                    labels = labels + '{0:s},'.format(data_str)
                else:
                    labels = labels + \
                        '{0:s} ({1:s}),'.format(data_str, unit_str)

            # Remove the comma at the end of the string and append a new-line character
            labels = labels[:-1]
            header = labels + '\n'
        else:
            header = ''

        self.log_file_rows[packet_type] += 1

        # Loop through the items in the data dictionary and append to an output string
        #   (with precision based on the data type defined in the json properties file)
        str = ''
        for i, (k, v) in enumerate(data.items()):
            output_packet_type = output_packet['payload'][i]['type']

            if output_packet['payload'][i].__contains__('scaling'):
                str +='{0},'.format(v)
            else:
                if output_packet_type == 'uint32' or output_packet_type == 'int32' or \
                output_packet_type == 'uint16' or output_packet_type == 'int16' or \
                output_packet_type == 'uint64' or output_packet_type == 'int64':
                    # integers and unsigned integers
                    str += '{0:d},'.format(v)
                elif output_packet_type == 'double':
                    # double
                    str += '{0:0.8f},'.format(v)  # 15.12
                elif output_packet_type == 'float':
                    str += '{0:0.4f},'.format(v)  # 12.8
                elif output_packet_type == 'uint8':
                    # byte
                    str += '{0:d},'.format(v)
                elif output_packet_type == 'uchar' or output_packet_type == 'char' or output_packet_type == 'string':
                    # character
                    str += '{:},'.format(v)
                else:
                    # unknown
                    str += '{0:3.5f},'.format(v)
        #
        str = header + str[:-1] + '\n'

        self.log_files_obj[packet_type].write(str)
        self.log_files_obj[packet_type].flush()

        if self.ws:
            self.data_lock.acquire()
            self.data_dict[self.log_files[packet_type]
                           ] = self.data_dict[self.log_files[packet_type]] + str
            self.data_lock.release()

    def set_info(self, info):
        self.device_log_info = info
        pass

    def set_user_id(self, user_id):
        self.user_id = user_id
        if not isinstance(self.user_id, str):
            self.user_id = str(self.user_id)

    def set_user_access_token(self, access_token):
        self.db_user_access_token = access_token

    def get_sas_token(self):
        try:
            url = self.host_url + "token/storagesas"
            headers = {'Content-type': 'application/json',
                       'Authorization': self.db_user_access_token}
            response = requests.post(url, headers=headers)
            rev = response.json()
            if 'token' in rev:
                self.sas_token = rev['token']
            else:
                self.sas_token = ''
                print('Error: Get sas token failed!')
            print('sas_token', self.sas_token)
        except Exception as e:
            print('Exception when get_sas_token:', e)

    def save_to_ans_platform(self, packet_type, file_name, url_name):
        ''' Upload CSV related information to the database.
        '''
        if not self.device_log_info:
            return False

        try:
            self.device_log_info['fileName'] = file_name
            self.device_log_info['url'] = url_name
            self.device_log_info['userId'] = self.user_id
            self.device_log_info['logInfo']['packetType'] = packet_type
            data = self.device_log_info

            # data = {
            #     "type": self.device_log_info['type'],
            #     "model": self.device_log_info['name'],
            #     "fileName": file_name,
            #     "url": file_name,
            #     "userId": self.user_id,
            #     "logInfo": {
            #             "pn": self.device_log_info['pn'],
            #             "sn": self.device_log_info['sn'],
            #             "packetType": packet_type,
            #             "insProperties": json.dumps(self.device_properties)
            #     }
            # }

            url = self.host_url + "api/recordLogs/post"
            data_json = json.dumps(data)
            headers = {'Content-type': 'application/json',
                       'Authorization': self.db_user_access_token}
            response = requests.post(url, data=data_json, headers=headers)
            return True if 'success' in response.json() else False
        except Exception as e:
            print('Exception when update db:', e)

    def internet_on(self):
        try:
            url = 'https://navview.blob.core.windows.net/'
            if sys.version_info[0] > 2:
                import urllib.request
                response = urllib.request.urlopen(url, timeout=1)
            else:
                import urllib2
                response = urllib2.urlopen(url, timeout=1)
            # print(response.read())
            return True
        except urllib2.URLError as err:
            return False



"""
Helper
"""
import struct
import sys
from .dict_extend import Dict

COMMAND_START = [0x55, 0x55]


def build_packet(message_type, message_bytes=[]):
    '''
    Build final packet
    '''
    packet = []
    packet.extend(bytearray(message_type, 'utf-8'))

    msg_len = len(message_bytes)
    packet.append(msg_len)
    final_packet = packet + message_bytes
    # print(message_type, final_packet)
    return COMMAND_START + final_packet + calc_crc(final_packet)


def build_input_packet(name, properties=None, param=False, value=False):
    '''
    Build input packet
    '''
    packet = []

    if not param and not value:
        packet = build_packet(name)
    else:
        payload = unpack_payload(name, properties, param, value)
        packet = build_packet(name, payload)
    return packet


def build_bootloader_input_packet(name, data_len=False, addr=False, data=False):
    '''
    Build bootloader input packet
    '''
    if not data_len and not addr and not data:
        packet = build_packet(name)
    else:
        payload = block_payload(data_len, addr, data)
        packet = build_packet(name, payload)
    return packet


def build_read_eeprom_input_packet(start, word_len):
    '''
    Build RE command
    '''
    payload = []
    payload.append((start & 0xFF00) >> 8)
    payload.append(start & 0x00FF)
    payload.append(word_len)
    packet = build_packet('RE', payload)
    return packet


def build_write_eeprom_input_packet(start, word_len, data):
    '''
    Build WE command
    '''
    name_bytes = list(struct.unpack('BB', bytearray('WE', 'utf-8')))
    payload = []
    payload.append((start & 0xFF00) >> 8)
    payload.append(start & 0x00FF)
    payload.append(word_len)
    payload.extend(data)
    command = COMMAND_START + name_bytes + [word_len*2+3] + payload
    packet = command + calc_crc(command[2:command[4]+5])
    return packet


def build_unlock_eeprom_packet(sn):
    '''
    Build UE command
    '''
    sn_crc = calc_crc(sn)
    payload = sn_crc
    packet = build_packet('UE', payload)
    return packet


def build_lock_eeprom_packet():
    '''
    Build UE command
    '''
    packet = build_packet('LE')
    return packet


def unpack_payload(name, properties, param=False, value=False):
    '''
    Unpack payload
    '''
    input_packet = next(
        (x for x in properties['userMessages']['inputPackets'] if x['name'] == name), None)

    if name == 'ma':
        input_action = next(
            (x for x in input_packet['inputPayload'] if x['actionName'] == param), None)
        return [input_action['actionID']]
    elif input_packet is not None:
        if input_packet['inputPayload']['type'] == 'paramId':
            return list(struct.unpack("4B", struct.pack("<L", param)))
        elif input_packet['inputPayload']['type'] == 'userParameter':
            payload = list(struct.unpack("4B", struct.pack("<L", param)))
            if properties['userConfiguration'][param]['type'] == 'uint64':
                payload += list(struct.unpack("8B", struct.pack("<Q", value)))
            elif properties['userConfiguration'][param]['type'] == 'int64':
                payload += list(struct.unpack("8B", struct.pack("<q", value)))
            elif properties['userConfiguration'][param]['type'] == 'double':
                payload += list(struct.unpack("8B",
                                              struct.pack("<d", float(value))))
            elif properties['userConfiguration'][param]['type'] == 'uint32':
                payload += list(struct.unpack("4B", struct.pack("<I", value)))
            elif properties['userConfiguration'][param]['type'] == 'int32':
                payload += list(struct.unpack("4B", struct.pack("<i", value)))
            elif properties['userConfiguration'][param]['type'] == 'float':
                payload += list(struct.unpack("4B", struct.pack("<f", value)))
            elif properties['userConfiguration'][param]['type'] == 'uint16':
                payload += list(struct.unpack("2B", struct.pack("<H", value)))
            elif properties['userConfiguration'][param]['type'] == 'int16':
                payload += list(struct.unpack("2B", struct.pack("<h", value)))
            elif properties['userConfiguration'][param]['type'] == 'uint8':
                payload += list(struct.unpack("1B", struct.pack("<B", value)))
            elif properties['userConfiguration'][param]['type'] == 'int8':
                payload += list(struct.unpack("1B", struct.pack("<b", value)))
            elif 'char' in properties['userConfiguration'][param]['type']:
                c_len = int(properties['userConfiguration']
                            [param]['type'].replace('char', ''))
                if isinstance(value, int):
                    length = len(str(value))
                    payload += list(struct.unpack('{0}B'.format(length),
                                                  bytearray(str(value), 'utf-8')))
                else:
                    length = len(value)
                    payload += list(struct.unpack('{0}B'.format(length),
                                                  bytearray(value, 'utf-8')))
                for i in range(c_len-length):
                    payload += [0x00]
            elif properties['userConfiguration'][param]['type'] == 'ip4':
                ip_address = value.split('.')
                ip_address_v4 = list(map(int, ip_address))
                for i in range(4):
                    payload += list(struct.unpack("1B",
                                                  struct.pack("<B", ip_address_v4[i])))
            elif properties['userConfiguration'][param]['type'] == 'ip6':
                ip_address = value.split('.')
                payload += list(struct.unpack('6B',
                                              bytearray(ip_address, 'utf-8')))

            return payload


def block_payload(data_len, addr, data):
    '''
    Block payload
    '''
    data_bytes = []
    addr_3 = (addr & 0xFF000000) >> 24
    addr_2 = (addr & 0x00FF0000) >> 16
    addr_1 = (addr & 0x0000FF00) >> 8
    addr_0 = (addr & 0x000000FF)
    data_bytes.insert(len(data_bytes), addr_3)
    data_bytes.insert(len(data_bytes), addr_2)
    data_bytes.insert(len(data_bytes), addr_1)
    data_bytes.insert(len(data_bytes), addr_0)
    data_bytes.insert(len(data_bytes), data_len)
    for i in range(data_len):
        if sys.version_info > (3, 0):
            data_bytes.insert(len(data_bytes), data[i])
        else:
            data_bytes.insert(len(data_bytes), ord(data[i]))
    return data_bytes


def calc_crc(payload):
    '''
    Calculates 16-bit CRC-CCITT
    '''
    crc = 0x1D0F
    for bytedata in payload:
        crc = crc ^ (bytedata << 8)
        i = 0
        while i < 8:
            if crc & 0x8000:
                crc = (crc << 1) ^ 0x1021
            else:
                crc = crc << 1
            i += 1

    crc = crc & 0xffff
    crc_msb = (crc & 0xFF00) >> 8
    crc_lsb = (crc & 0x00FF)
    return [crc_msb, crc_lsb]


def clear_elements(list_instance):
    '''
    clear list
    '''
    if sys.version_info < (3, 0):
        list_instance[:] = []
    else:
        list_instance.clear()


def dict_to_object(dict_obj):
    '''
    Convert Dict to Object
    '''
    if not isinstance(dict_obj, dict):
        return dict_obj
    inst = Dict()
    for key, val in dict_obj.items():
        inst[key] = dict_to_object(val)
    return inst


def name_convert_camel_to_snake(camel_name):
    '''
    Convert Camel naming to snake case
    '''
    chars = []
    underscore = '_'

    lower_camel_name = camel_name.lower()

    for i, char in enumerate(camel_name):
        add_underscore = False
        lower_char = lower_camel_name[i]
        if char != lower_char:
            add_underscore = True if i > 0 else False

        if add_underscore:
            chars.append(underscore+lower_char)
        else:
            chars.append(lower_char)

    return ''.join(chars)



import os
import logging
from logging import handlers


class MessageStoreHandler(logging.Handler):
    def __init__(self, store):
        logging.Handler.__init__(self)
        self._store = store

    def mapLogRecord(self, record):
        return record.__dict__

    def emit(self, record):
        msg = self.mapLogRecord(record)
        self._store.append(msg)


class AppLogger(object):
    level_relations = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'crit': logging.CRITICAL
    }

    def __init__(self, filename, level='info', when='D',
                 backCount=1, gen_file=False, console_log=False,
                 fmt='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'):
        self._store = None
        self.logger = logging.getLogger()
        self.format_str = logging.Formatter(fmt)
        self.logger.setLevel(self.level_relations.get(level))
        if console_log:
            console_output = logging.StreamHandler()
            console_output.setFormatter(self.format_str)
            self.logger.addHandler(console_output)

        if gen_file:
            folder_path = os.path.dirname(filename)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
            daily_file_output = handlers.TimedRotatingFileHandler(
                filename=filename, when=when, backupCount=backCount, encoding='utf-8')
            daily_file_output.setFormatter(self.format_str)
            self.logger.addHandler(daily_file_output)

    def enable_msg_store_handler(self, store):
        if not self._store:
            self._store = MessageStoreHandler(store)
            self._store.setFormatter(self.format_str)
            self.logger.addHandler(self._store)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, **kwargs)
